/* alf.c
 * 
 * Copyright (C) 2009 Patrick Alken
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "alf.h"

/*
 * The routines in this module compute associated Legendre functions
 * (ALFs) up to order and degree 2700, using the method described
 * in
 *
 * [1] S. A. Holmes and W. E. Featherstone, A unified approach
 *     to the Clenshaw summation and the recursive computation of very
 *     high degree and order normalised associated Legendre functions,
 *     Journal of Geodesy, 76, pg. 279-299, 2002.
 *
 * Further information on ALFs can be found in
 *
 * [2] Abramowitz and Stegun, Handbook of Mathematical Functions,
 *     Chapter 8, 1972.
 */

static int alf_Plm_array_none(const int lmax, const double x,
                              double result_array[], alf_workspace *w);
static int alf_Plm_deriv_array_none(const int lmax, const double x,
                                    double result_array[],
                                    double result_deriv_array[],
                                    alf_workspace *w);
static int alf_Plm_array_norm(const int lmax, const double x,
                              double result_array[], alf_workspace *w);
static int alf_Plm_deriv_array_norm(const int lmax, const double x,
                                    double result_array[],
                                    double result_deriv_array[],
                                    alf_workspace *w);
static int alf_Plm_deriv2_array_norm(const int lmax, const double x,
                                     double result_array[],
                                     double result_deriv_array[],
                                     double result_deriv2_array[],
                                     alf_workspace *w);

/*
alf_alloc()
  Allocate an ALF workspace

Inputs: lmax - maximum degree to be computed

Return: pointer to workspace
*/

alf_workspace *
alf_alloc(const int lmax)
{
  alf_workspace *w;
  int l;

  if (lmax < 0)
    {
      printf("lmax must be >= 0\n");
      return 0;
    }

  w = calloc(1, sizeof(alf_workspace));
  if (!w)
    {
      printf("error allocating workspace\n");
      return 0;
    }

  w->sqrts = malloc(sizeof(double) * (2 * lmax + 2));
  if (w->sqrts == 0)
    {
      printf("error allocating sqrts\n");
      alf_free(w);
      return 0;
    }

  w->alpha = malloc(sizeof(double) * (lmax + 1));
  if (w->alpha == 0)
    {
      printf("error allocating alpha\n");
      alf_free(w);
      return 0;
    }

  w->lmax = lmax;
  w->csphase = 0.0;

  /*
   * precompute square roots of integers that are used several times
   * sqrts[i] = sqrt(i)
   */
  for (l = 0; l <= 2 * w->lmax + 1; ++l)
    w->sqrts[l] = sqrt((double) l);

  alf_params(ALF_CSPHASE_DEFAULT, ALF_CNORM_DEFAULT, ALF_NORM_DEFAULT, w);

  return w;
} /* alf_alloc() */

/*
alf_free()
  Free an ALF workspace
*/

void
alf_free(alf_workspace *w)
{
  if (w->sqrts)
    free(w->sqrts);

  if (w->alpha)
    free(w->alpha);

  free(w);
} /* alf_free() */

/*
alf_params()
  Define various parameters for the computation of associated
Legendre polynomials.

Inputs:
  csphase - 1 to apply the Condon-Shortley phase of (-1)^m
            0 to omit this phase
  cnorm   - 1 to use complex normalization
            0 to use real normalization
  norm    - type of normalization to use for ALFs (ALF_NORM_xxx)
  w       - workspace
*/

int
alf_params(const int csphase, const int cnorm, const alf_norm_t norm,
           alf_workspace *w)
{
  int l;

  if (csphase)
    w->csphase = -1.0;
  else
    w->csphase = 1.0;

  w->cnorm = cnorm;
  w->norm = norm;

  if (norm == ALF_NORM_SPHARM)
    {
      /* Y(l,m) = sqrt((2*l + 1) / (4 pi)) S(l,m) */
      for (l = 0; l <= w->lmax; ++l)
        w->alpha[l] = w->sqrts[2 * l + 1] / sqrt(4.0 * M_PI);
    }
  else if (norm == ALF_NORM_SCHMIDT)
    {
      for (l = 0; l <= w->lmax; ++l)
        w->alpha[l] = 1.0;
    }
  else if (norm == ALF_NORM_ORTHO)
    {
      /* N(l,m) = sqrt((2*l + 1) / 2) S(l,m) */
      for (l = 0; l <= w->lmax; ++l)
        w->alpha[l] = w->sqrts[2 * l + 1] / sqrt(2.0);
    }

  return 0;
} /* alf_params() */

/*
alf_Plm_array()

  Compute associated Legendre functions

Inputs: lmax         - maximum degree to compute
        x            - argument to ALFs
        result_array - where to store P(l,m) values
        w            - workspace
*/

int
alf_Plm_array(const int lmax, const double x, double result_array[],
              alf_workspace *w)
{
  if (lmax > w->lmax)
    {
      printf("alf_Plm_array: error: lmax is higher than previously specified lmax\n");
      return 1;
    }
  else if (fabs(x) > 1.0)
    {
      printf("alf_Plm_array: error: |x| must be <= 1: %f\n", x);
      return 1;
    }
  else
    {
      int s;

      if (w->norm == ALF_NORM_NONE)
        s = alf_Plm_array_none(lmax, x, result_array, w);
      else
        s = alf_Plm_array_norm(lmax, x, result_array, w);
        
      return s;
    }
} /* alf_Plm_array() */

/*
alf_Plm_deriv_array()

  Compute associated Legendre functions and their first derivatives

Inputs: lmax               - maximum degree to compute
        x                  - argument to ALFs
        result_array       - where to store P(l,m) values
        result_deriv_array - where to store dP(l,m)/dx values
        w                  - workspace
*/

int
alf_Plm_deriv_array(const int lmax, const double x, double result_array[],
                    double result_deriv_array[], alf_workspace *w)
{
  int s;

  if (lmax > w->lmax)
    {
      printf("alf_Plm_deriv_array: error: lmax is higher than previously specified lmax\n");
      return 1;
    }
  else if (fabs(x) > 1.0)
    {
      printf("alf_Plm_deriv_array: error: |x| must be <= 1: %f\n", x);
      return 1;
    }
  else if (fabs(x) == 1.0)
    {
      printf("alf_Plm_deriv_array: warning: derivative cannot be calculated at x = 1 or -1, computing Legendre functions only\n");
      s = alf_Plm_array(lmax, x, result_array, w);
      return s;
    }
  else
    {

      if (w->norm == ALF_NORM_NONE)
        s = alf_Plm_deriv_array_none(lmax, x, result_array,
                                     result_deriv_array, w);
      else
        s = alf_Plm_deriv_array_norm(lmax, x, result_array,
                                     result_deriv_array, w);
        
      return s;
    }
} /* alf_Plm_deriv_array() */

/*
alf_Plm_deriv2_array()

  Compute associated Legendre functions and their first and second
derivatives

Inputs: lmax                - maximum degree to compute
        x                   - argument to ALFs
        result_array        - where to store P(l,m) values
        result_deriv_array  - where to store dP(l,m)/dx values
        result_deriv2_array - where to store d^2P(l,m)/dx^2 values
        w                   - workspace
*/

int
alf_Plm_deriv2_array(const int lmax, const double x, double result_array[],
                     double result_deriv_array[],
                     double result_deriv2_array[], alf_workspace *w)
{
  if (lmax > w->lmax)
    {
      printf("alf_Plm_deriv2_array: error: lmax is higher than previously specified lmax\n");
      return 1;
    }
  else if (fabs(x) > 1.0)
    {
      printf("alf_Plm_deriv2_array: error: |x| must be <= 1: %f\n", x);
      return 1;
    }
  else if (fabs(x) == 1.0)
    {
      printf("alf_Plm_deriv2_array: error: derivative cannot be calculated at x = 1 or -1\n");
      return 1;
    }
  else
    {
      int s;

      if (w->norm == ALF_NORM_NONE)
        s = alf_Plm_deriv_array_none(lmax, x, result_array,
                                     result_deriv_array, w);
      else
        s = alf_Plm_deriv2_array_norm(lmax, x, result_array,
                                      result_deriv_array,
                                      result_deriv2_array, w);
        
      return s;
    }
} /* alf_Plm_deriv2_array() */

/*
alf_array_size()
  This routine returns the minimum result_array[] size needed
for a given lmax
*/

size_t
alf_array_size(const int lmax)
{
  return (size_t) ((lmax + 1) * (lmax + 2) / 2);
} /* alf_array_size() */

/*
alf_array_index()
This routine computes the index into a result_array[] corresponding
to a given (l,m)
*/

size_t
alf_array_index(const int l, const int m)
{
  return (l * (l + 1) / 2 + m);
} /* alf_index() */

/*********************************************************
 *                 INTERNAL ROUTINES                     *
 *********************************************************/

/*
alf_Plm_array_none()
  This routine computes unnormalized associated Legendre polynomials
*/

static int
alf_Plm_array_none(const int lmax, const double x,
                   double result_array[], alf_workspace *w)
{
  const double u = sqrt((1.0 - x) * (1.0 + x)); /* sin(theta) */
  int l, m;
  size_t k, kstart;
  double plm, pmm;
  double pm1,    /* P(l-1,m) */
         pm2;    /* P(l-2,m) */
  double twomm1; /* 2*m - 1 */

  /* initial values P(0,0) and P(1,0) */

  pm2 = 1.0; /* P(0,0) */
  pm1 = x;   /* P(1,0) */

  result_array[0] = pm2;

  if (lmax == 0)
    return 0;

  result_array[1] = pm1;

  /* Compute P(l,0) */

  k = 1;
  for (l = 2; l <= lmax; ++l)
    {
      k += l;
      plm = ((2*l - 1) * x * pm1 - (l - 1) * pm2) / (double) l;
      result_array[k] = plm;
      pm2 = pm1;
      pm1 = plm;
    }

  /* Compute P(m,m), P(m+1,m) and P(l,m) */

  pmm = 1.0;
  twomm1 = -1.0; /* 2 * m - 1 */

  kstart = 0;

  for (m = 1; m <= lmax - 1; ++m)
    {
      /*
       * compute
       *
       * P(m,m) = u * (2m - 1) P(m-1,m-1)
       * and
       * dP(m,m)/dx = -m * x * P(m,m) / u^2
       */
      kstart += m + 1;
      twomm1 += 2.0;
      pmm *= w->csphase * u * twomm1;
      result_array[kstart] = pmm;
      pm2 = pmm;

      /*
       * compute
       *
       * P(m+1,m) = (2 * m + 1) * x * P(m,m)
       * and
       * dP(m+1,m)/dx = [(2*m + 1) * P(m,m) - (m+1) * x * P(m+1,m)]/u^2
       */
      k = kstart + m + 1;
      pm1 = x * pmm * (2*m + 1);
      result_array[k] = pm1;

      /* compute P(l,m) */
      for (l = m + 2; l <= lmax; ++l)
        {
          k += l;
          plm = ((2*l - 1) * x * pm1 - (l + m - 1) * pm2) /
                (double) (l - m);
          result_array[k] = plm;
          pm2 = pm1;
          pm1 = plm;
        }
    } /* for (m = 1; m <= lmax - 1; ++m) */

  /* compute P(lmax,lmax) */

  kstart += m + 1;
  twomm1 += 2.0;
  pmm *= w->csphase * u * twomm1;
  result_array[kstart] = pmm;

  return 0;
} /* alf_Plm_array_none() */

/*
alf_Plm_deriv_array_none()
  This routine computes unnormalized associated Legendre polynomials
and their first derivatives.
*/

static int
alf_Plm_deriv_array_none(const int lmax, const double x,
                         double result_array[],
                         double result_deriv_array[], alf_workspace *w)
{
  const double u = sqrt((1.0 - x) * (1.0 + x)); /* sin(theta) */
  const double u_sq = u * u;
  int l, m;
  size_t k, kstart;
  double plm, pmm;
  double pm1,    /* P(l-1,m) */
         pm2;    /* P(l-2,m) */
  double twomm1; /* 2*m - 1 */

  /* initial values P(0,0) and P(1,0) */

  pm2 = 1.0; /* P(0,0) */
  pm1 = x;   /* P(1,0) */

  result_array[0] = pm2;
  result_deriv_array[0] = 0.0;

  if (lmax == 0)
    return 0;

  result_array[1] = pm1;
  result_deriv_array[1] = 1.0;

  /* Compute P(l,0) */

  k = 1;
  for (l = 2; l <= lmax; ++l)
    {
      k += l;
      plm = ((2*l - 1) * x * pm1 - (l - 1) * pm2) / (double) l;
      result_array[k] = plm;
      result_deriv_array[k] = l * (pm1 - x * plm) / u_sq;
      pm2 = pm1;
      pm1 = plm;
    }

  /* Compute P(m,m), P(m+1,m) and P(l,m) */

  pmm = 1.0;
  twomm1 = -1.0; /* 2 * m - 1 */

  kstart = 0;

  for (m = 1; m <= lmax - 1; ++m)
    {
      /*
       * compute
       *
       * P(m,m) = u * (2m - 1) P(m-1,m-1)
       * and
       * dP(m,m)/dx = -m * x * P(m,m) / u^2
       */
      kstart += m + 1;
      twomm1 += 2.0;
      pmm *= w->csphase * u * twomm1;
      result_array[kstart] = pmm;
      result_deriv_array[kstart] = - m * x * pmm / u_sq;
      pm2 = pmm;

      /*
       * compute
       *
       * P(m+1,m) = (2 * m + 1) * x * P(m,m)
       * and
       * dP(m+1,m)/dx = [(2*m + 1) * P(m,m) - (m+1) * x * P(m+1,m)]/u^2
       */
      k = kstart + m + 1;
      pm1 = x * pmm * (2*m + 1);
      result_array[k] = pm1;
      result_deriv_array[k] = ((2*m + 1) * pmm - (m + 1) * x * pm1) / u_sq;

      /* compute P(l,m) */
      for (l = m + 2; l <= lmax; ++l)
        {
          k += l;
          plm = ((2*l - 1) * x * pm1 - (l + m - 1) * pm2) /
                (double) (l - m);
          result_array[k] = plm;
          result_deriv_array[k] = ((l + m) * pm1 - l * x * plm) / u_sq;
          pm2 = pm1;
          pm1 = plm;
        }
    } /* for (m = 1; m <= lmax - 1; ++m) */

  /* compute P(lmax,lmax) */

  kstart += m + 1;
  twomm1 += 2.0;
  pmm *= w->csphase * u * twomm1;
  result_array[kstart] = pmm;
  result_deriv_array[kstart] = - lmax * x * pmm / u_sq;

  return 0;
} /* alf_Plm_deriv_array_none() */

/*
alf_Plm_array_norm()
  This routine computes Schmidt semi-normalized associated
Legendre polynomials and then scales them to other normalizations
according to the precomputed scale factors in w->alpha
*/

static int
alf_Plm_array_norm(const int lmax, const double x,
                   double result_array[], alf_workspace *w)
{
  const double scalef = 1.0e-280;
  const double u = sqrt((1.0 - x) * (1.0 + x)); /* sin(theta) */
  int l, m;
  size_t k, kstart;
  double plm, pmm;
  double rescalem;
  double pm1, /* S(l-1,m) */
         pm2; /* S(l-2,m) */

  /* initial values S(0,0) and S(1,0) */

  pm2 = 1.0; /* S(0,0) */
  pm1 = x;   /* S(1,0) */

  result_array[0] = pm2 * w->alpha[0];

  if (lmax == 0)
    return 0;

  result_array[1] = pm1 * w->alpha[1];

  /* Compute S(l,0), no scaling required */

  k = 1;
  for (l = 2; l <= lmax; ++l)
    {
      k += l;
      plm = ((2*l - 1) * x * pm1 - (l - 1) * pm2) / (double) l;
      result_array[k] = w->alpha[l] * plm;
      pm2 = pm1;
      pm1 = plm;
    }

  /* Compute S(m,m), S(m+1,m) and S(l,m) */

  if (w->cnorm)
    pmm = scalef;
  else
    pmm = sqrt(2.0) * scalef;

  rescalem = 1.0 / scalef;
  kstart = 0;

  for (m = 1; m <= lmax - 1; ++m)
    {
      rescalem *= u;

      /*
       * compute
       *
       * S(m,m) = -u * sqrt((2m - 1) / (2m)) S(m-1,m-1)
       * and
       * dS(m,m)/dx = -m * x * S(m,m) / u^2
       */
      kstart += m + 1;
      pmm *= w->csphase * w->sqrts[2 * m + 1] / w->sqrts[2 * m];
      result_array[kstart] = w->alpha[m] * pmm * rescalem /
                             w->sqrts[2 * m + 1];
      pm2 = pmm / w->sqrts[2 * m + 1];

      /*
       * compute
       *
       * S(m+1,m) = sqrt((2 * m + 1)/2) * x * S(m,m)
       * and
       * dY(m+1,m)/dx = [sqrt(2*m+3) * Y(m,m) - (m+1) * x * Y(m+1,m)]/u^2
       */
      k = kstart + m + 1;
      pm1 = x * w->sqrts[2 * m + 1] * pm2;
      result_array[k] = w->alpha[m + 1] * pm1 * rescalem;

      /* compute S(l,m) */
      for (l = m + 2; l <= lmax; ++l)
        {
          k += l;
          plm =
            (2*l - 1) / w->sqrts[l + m] / w->sqrts[l - m] * x * pm1 -
            w->sqrts[l - m - 1] * w->sqrts[l + m - 1] /
            w->sqrts[l + m] / w->sqrts[l - m] * pm2;
          result_array[k] = w->alpha[l] * plm * rescalem;
          pm2 = pm1;
          pm1 = plm;
        }
    } /* for (m = 1; m <= lmax - 1; ++m) */

  /* compute S(lmax,lmax) */

  rescalem *= u;
  kstart += m + 1;
  pmm *= w->csphase / w->sqrts[2 * lmax];
  result_array[kstart] = w->alpha[lmax] * pmm * rescalem;

  return 0;
} /* alf_Plm_array_norm() */

/*
alf_Plm_deriv_array_norm()
  This routine computes Schmidt semi-normalized associated
Legendre polynomials and their first derivatives and then scales them
to other normalizations according to the precomputed scale factors in
w->alpha
*/

static int
alf_Plm_deriv_array_norm(const int lmax, const double x,
                         double result_array[],
                         double result_deriv_array[], alf_workspace *w)
{
  const double scalef = 1.0e-280;
  const double u = sqrt((1.0 - x) * (1.0 + x)); /* sin(theta) */
  const double u_sq = u * u;
  int l, m;
  size_t k, kstart;
  double plm, pmm;
  double rescalem;
  double pm1, /* S(l-1,m) */
         pm2; /* S(l-2,m) */

  /* initial values S(0,0) and S(1,0) */

  pm2 = 1.0; /* S(0,0) */
  pm1 = x;   /* S(1,0) */

  result_array[0] = pm2 * w->alpha[0];
  result_deriv_array[0] = 0.0;

  if (lmax == 0)
    return 0;

  result_array[1] = pm1 * w->alpha[1];
  result_deriv_array[1] = w->alpha[1];

  /* Compute S(l,0), no scaling required */

  k = 1;
  for (l = 2; l <= lmax; ++l)
    {
      k += l;
      plm = ((2*l - 1) * x * pm1 - (l - 1) * pm2) / (double) l;
      result_array[k] = w->alpha[l] * plm;
      result_deriv_array[k] = w->alpha[l] * l * (pm1 - x * plm) / u_sq;
      pm2 = pm1;
      pm1 = plm;
    }

  /*
   * Compute S(m,m), S(m+1,m) and S(l,m)
   *
   * The variable 'pmm' holds the quantity:
   *
   * pmm = scalef * S(m,m) / u^m = scalef * Product_{i=1,m} a_i S(0,0)
   *
   * where
   *
   * a_i = sqrt((2m+1)/(2m))
   *
   * and S(0,0) = 1
   */

  if (w->cnorm)
    pmm = scalef;
  else
    pmm = sqrt(2.0) * scalef;

  rescalem = 1.0 / scalef;
  kstart = 0;

  for (m = 1; m <= lmax - 1; ++m)
    {
      /* rescalem = u^m / scalef */
      rescalem *= u;

      /*
       * compute
       *
       * S(m,m) = -u * sqrt((2m - 1) / (2m)) S(m-1,m-1)
       * and
       * dS(m,m)/dx = -m * x * S(m,m) / u^2
       *
       * use rescalem to undo the scaling when storing the final
       * result
       */
      kstart += m + 1;
      pmm *= w->csphase * w->sqrts[2 * m + 1] / w->sqrts[2 * m];
      result_array[kstart] = w->alpha[m] * pmm * rescalem /
                             w->sqrts[2 * m + 1];
      result_deriv_array[kstart] = - m * x *
                                   result_array[kstart] / u_sq;
      pm2 = pmm / w->sqrts[2 * m + 1];

      /*
       * compute
       *
       * S(m+1,m) = sqrt((2 * m + 1)/2) * x * S(m,m)
       * and
       * dS(m+1,m)/dx = [sqrt(2*m+3) * S(m,m) - (m+1) * x * S(m+1,m)]/u^2
       */
      k = kstart + m + 1;
      pm1 = x * w->sqrts[2 * m + 1] * pm2;
      result_array[k] = w->alpha[m + 1] * pm1 * rescalem;
      result_deriv_array[k] =
        w->alpha[m + 1] *
        ((pm2 * rescalem) * w->sqrts[2 * m + 1] -
         x * (m + 1) * (pm1 * rescalem)) / u_sq;

      /* compute S(l,m) */
      for (l = m + 2; l <= lmax; ++l)
        {
          k += l;
          plm =
            (2*l - 1) / w->sqrts[l + m] / w->sqrts[l - m] * x * pm1 -
            w->sqrts[l - m - 1] * w->sqrts[l + m - 1] /
            w->sqrts[l + m] / w->sqrts[l - m] * pm2;
          result_array[k] = w->alpha[l] * plm * rescalem;
          result_deriv_array[k] =
            w->alpha[l] *
            (w->sqrts[l + m] * w->sqrts[l - m] * (pm1 * rescalem) -
             l * x * (plm * rescalem)) / u_sq;
          pm2 = pm1;
          pm1 = plm;
        }
    } /* for (m = 1; m <= lmax - 1; ++m) */

  /* compute S(lmax,lmax) */

  rescalem *= u;
  kstart += m + 1;
  pmm *= w->csphase / w->sqrts[2 * lmax];
  result_array[kstart] = w->alpha[lmax] * pmm * rescalem;
  result_deriv_array[kstart] = - lmax * x * result_array[kstart] / u_sq;

  return 0;
} /* alf_Plm_deriv_array_norm() */

/*
alf_Plm_deriv2_array_norm()
  This routine computes Schmidt semi-normalized associated
Legendre polynomials and their first and second derivatives and then
scales them to other normalizations according to the precomputed scale
factors in w->alpha
*/

static int
alf_Plm_deriv2_array_norm(const int lmax, const double x,
                          double result_array[],
                          double result_deriv_array[],
                          double result_deriv2_array[],
                          alf_workspace *w)
{
  const double scalef = 1.0e-280;
  const double u = sqrt((1.0 - x) * (1.0 + x)); /* sin(theta) */
  const double u_sq = u * u;
  int l, m;
  size_t k, kstart;
  double plm, pmm;
  double dplm, d2plm;
  double rescalem;
  double pm1,   /* S(l-1,m) */
         pm2;   /* S(l-2,m) */
  double dpm1,  /* S'(l-1,m) */
         dpm2;  /* S'(l-2,m) */
  double d2pm1, /* S''(l-1,m) */
         d2pm2; /* S''(l-2,m) */

  /* initial values S(0,0) and S(1,0) */

  pm2 = 1.0; /* S(0,0) */
  pm1 = x;   /* S(1,0) */

  dpm2 = 0.0; /* S'(0,0) */
  dpm1 = 1.0; /* S'(1,0) */

  d2pm2 = 0.0; /* S''(0,0) */
  d2pm1 = 0.0; /* S''(1,0) */

  result_array[0] = pm2 * w->alpha[0];
  result_deriv_array[0] = dpm2 * w->alpha[0];
  result_deriv2_array[0] = d2pm2 * w->alpha[0];

  if (lmax == 0)
    return 0;

  result_array[1] = pm1 * w->alpha[1];
  result_deriv_array[1] = dpm1 * w->alpha[1];
  result_deriv2_array[1] = d2pm1 * w->alpha[1];

  /* Compute S(l,0), no scaling required */

  k = 1;
  for (l = 2; l <= lmax; ++l)
    {
      double c1 = (2.0*l - 1) / (double) l;
      double c2 = (l - 1.0) / (double) l;

      k += l;

      plm = c1 * x * pm1 - c2 * pm2;
      result_array[k] = w->alpha[l] * plm;

      dplm = c1 * (pm1 + x * dpm1) - c2 * dpm2;
      result_deriv_array[k] = w->alpha[l] * dplm;

      d2plm = c1 * (2.0 * dpm1 + x * d2pm1) - c2 * d2pm2;
      result_deriv2_array[k] = w->alpha[l] * d2plm;

      pm2 = pm1;
      pm1 = plm;

      dpm2 = dpm1;
      dpm1 = dplm;

      d2pm2 = d2pm1;
      d2pm1 = d2plm;
    }

  /*
   * Compute S(m,m), S(m+1,m) and S(l,m)
   *
   * The variable 'pmm' holds the quantity:
   *
   * pmm = scalef * S(m,m) / u^m = scalef * Product_{i=1,m} a_i S(0,0)
   *
   * where
   *
   * a_i = sqrt((2m-1)/(2m))
   *
   * and S(0,0) = 1
   */

  if (w->cnorm)
    pmm = scalef;
  else
    pmm = sqrt(2.0) * scalef;

  rescalem = 1.0 / scalef;
  kstart = 0;

  for (m = 1; m <= lmax - 1; ++m)
    {
      /* rescalem = u^m / scalef */
      rescalem *= u;

      /*
       * compute
       *
       * S(m,m)/u^m = sqrt((2m - 1) / (2m)) S(m-1,m-1) / u^(m-1)
       * and
       * dS(m,m)/dx = -m * x * S(m,m) / u^2
       *
       * use rescalem to undo the scaling when storing the final
       * result
       */
      kstart += m + 1;
      pmm *= w->csphase * w->sqrts[2 * m - 1] / w->sqrts[2 * m];
      result_array[kstart] = w->alpha[m] * (pmm * rescalem);
      result_deriv_array[kstart] = - m * x *
                                   result_array[kstart] / u_sq;
      result_deriv2_array[kstart] = (2.0 * x * result_deriv_array[kstart] +
                                     (m * m / u_sq - m * (m + 1.0)) *
                                     result_array[kstart]) / u_sq;
      pm2 = pmm;

      /*
       * compute
       *
       * S(m+1,m) = sqrt(2 * m + 1) * x * S(m,m)
       * and
       * dS(m+1,m)/dx = [sqrt(2*m+1) * S(m,m) - (m+1) * x * S(m+1,m)]/u^2
       */
      k = kstart + m + 1;
      pm1 = x * w->sqrts[2 * m + 1] * pmm;
      result_array[k] = w->alpha[m + 1] * (pm1 * rescalem);
      result_deriv_array[k] =
        w->alpha[m + 1] *
        ((pmm * rescalem) * w->sqrts[2 * m + 1] -
         x * (m + 1) * (pm1 * rescalem)) / u_sq;
      result_deriv2_array[k] = (2.0 * x * result_deriv_array[k] +
                                (m * m / u_sq - (m + 1)*(m + 2)) *
                                result_array[k]) / u_sq;

      /* compute S(l,m) */
      for (l = m + 2; l <= lmax; ++l)
        {
          k += l;
          plm =
            (2*l - 1) / w->sqrts[l + m] / w->sqrts[l - m] * x * pm1 -
            w->sqrts[l - m - 1] * w->sqrts[l + m - 1] /
            w->sqrts[l + m] / w->sqrts[l - m] * pm2;
          result_array[k] = w->alpha[l] * (plm * rescalem);
          result_deriv_array[k] =
            w->alpha[l] *
            (w->sqrts[l + m] * w->sqrts[l - m] * (pm1 * rescalem) -
             l * x * (plm * rescalem)) / u_sq;
          result_deriv2_array[k] = (2.0 * x * result_deriv_array[k] +
                                    (m * m / u_sq - l * (l + 1.0)) *
                                    result_array[k]) / u_sq;
          pm2 = pm1;
          pm1 = plm;
        }
    } /* for (m = 1; m <= lmax - 1; ++m) */

  /* compute S(lmax,lmax) */

  rescalem *= u;
  kstart += m + 1;
  pmm *= w->csphase * w->sqrts[2 * lmax - 1] / w->sqrts[2 * lmax];
  result_array[kstart] = w->alpha[lmax] * (pmm * rescalem);
  result_deriv_array[kstart] = - lmax * x * result_array[kstart] / u_sq;
  result_deriv2_array[kstart] = (2.0 * x * result_deriv_array[kstart] +
                                 (lmax * lmax / u_sq - lmax*(lmax + 1.0)) *
                                 result_array[kstart]) / u_sq;

  return 0;
} /* alf_Plm_deriv2_array_norm() */
